#include <cmath>
#include <string>
#include <fstream>
#include <streambuf>
#include <algorithm>

#include <libreborn/libreborn.h>
#include <symbols/minecraft.h>

#include <mods/init/init.h>
#include <mods/misc/misc.h>
#include <mods/chat/chat.h>
#include <mods/server/server.h>
#include <mods/feature/feature.h>

bool compat_mode = true;
static std::string getBlocks(CommandServer *commandserver, Vec3 start, Vec3 end) {
    int startx = start.x, starty = start.y, startz = start.z;
    int endx = end.x, endy = end.y, endz = end.z;

    // Offset cords
    commandserver->pos_translator.from(startx, starty, startz);
    commandserver->pos_translator.from(endx, endy, endz);

    if (endx < startx) std::swap(startx, endx);
    if (endy < starty) std::swap(starty, endy);
    if (endz < startz) std::swap(startz, endz);

    // Get
    std::string ret = "";
    for (int x = startx; x <= endx; x++) {
        for (int y = starty; y <= endy; y++) {
            for (int z = startz; z <= endz; z++) {
                ret += std::to_string(commandserver->minecraft->level->getTile(x, y, z)) + ",";
            }
        }
    }
    if (ret.size() > 1) ret.pop_back();
    return ret + "\n";
}

static void setDir(Entity *entity, float x, float y, float z) {
    if (entity == NULL) return;

    constexpr double _2PI = 2 * M_PI;
    if (x == 0 && z == 0) {
        entity->pitch = y > 0 ? -90 : 90;
        return;
    }

    entity->yaw = fmod((float) atan2(-x, z), (float) _2PI) * (180.0f / M_PI);
    float xz = sqrt(x * x + z * z);
    entity->pitch = (float) atan(-y / xz) * (180.0f / M_PI);
}
static Vec3 getDir(Entity *entity) {
    float y  = -sin(entity->pitch * (M_PI / 180));
    float xz =  cos(entity->pitch * (M_PI / 180));
    float x = -xz * sin(entity->yaw * (M_PI / 180));
    float z =  xz * cos(entity->yaw * (M_PI / 180));
    return Vec3{x, y, z};
}

static std::string getEntityData(CommandServer *commandserver, Entity *entity) {
    float x = entity->x, y = entity->y - entity->height_offset, z = entity->z;
    commandserver->pos_translator.to(x, y, z);
    return std::to_string(entity->id) + "," +
        // type
        std::to_string(entity->getEntityTypeId()) + "," +
        // name
        misc_get_entity_name(entity) + "," +
        // x
        std::to_string(x) + "," +
        // y
        std::to_string(y) + "," +
        // z
        std::to_string(z);
}

static std::string replace(std::string str, std::string s1, std::string s2) {
    size_t i = str.find(s1, 0);
    while (i != std::string::npos) {
        str.replace(i, s1.length(), s2);
        i = str.find(s1, i + s2.length());
    }
    return str;
}

static float distance_between(Entity *e1, Entity *e2) {
    if (e1 == NULL || e2 == NULL) return 0;
    float dx = e2->x - e1->x;
    float dy = e2->y - e1->y;
    float dz = e2->z - e1->z;
    return sqrt(dx * dx + dy * dy + dz * dz);
}

struct ProjectileHitEvent {
    int x, y, z;
    std::string owner = "";
    int targetId = 0;

    std::string toString(CommandServer *commandserver) {
        float nx = x, ny = y, nz = z;
        commandserver->pos_translator.to(nx, ny, nz);
        return std::to_string((int) nx) + ","
             + std::to_string((int) ny) + ","
             + std::to_string((int) nz) + ","
             + (compat_mode ? "1," : "")
             + owner + "," + std::to_string(targetId);
    }
};

struct ChatEvent {
    std::string message = "";
    bool from_player = false;

    std::string toString() {
        return message;
    }
};

constexpr int EVENT_SIZE = 50;
static int hit_events_at = 0, hit_events_start = 0;
static ProjectileHitEvent hitEvents[EVENT_SIZE];
static int chat_events_at = 0, chat_events_start = 0;
static ChatEvent chatEvents[EVENT_SIZE];

template <typename T>
static void push_circular_queue(T event, T events[], int &start, int &at) {
    events[at] = event;
    at++;
    at %= EVENT_SIZE;
    if (at == start) {
        start++;
        start %= EVENT_SIZE;
    }
}

template <typename T>
static T pop_circular_queue(T events[], int &start, int &at) {
    if (start == at) {
        ERR("Over popped projectile hit events!");
    }
    T ret = events[start];
    start++;
    start %= EVENT_SIZE;
    return ret;
}

static const std::string fail = "Fail\n";
std::unordered_map<int, int> modern_entity_id_mapping = {
    {93, 10}, /* Chicken */
    {92, 11}, /* Cow */
    {90, 12}, /* Pig */
    {91, 13}, /* Sheep */
    {54, 32}, /* Zombie */
    {50, 33}, /* Creeper */
    {51, 34}, /* Skeleton */
    {52, 35}, /* Spider */
    {57, 36}, /* PigZombie */
    /* ItemEntity isn't available */
    {20, 65}, /* PrimedTnt */
    /* FallingTile isn't available */
    {10, 80}, /* Arrow */
    {11, 81}, /* Snowball */
    {7, 82}, /* Egg */
    {9, 83}, /* Painting */
};
std::string CommandServer_parse_injection(CommandServer_parse_t old, CommandServer *commandserver, ConnectedClient &client, const std::string &command) {
    size_t arg_start = command.find("(");
    if (arg_start == std::string::npos) return fail;
    std::string cmd = command.substr(0, arg_start);

    // rfind so ) in input doesn't break
    size_t cmd_end = command.rfind(")");
    if (cmd_end == std::string::npos) return fail;
    std::string args = command.substr(arg_start + 1, cmd_end - arg_start - 1);

    // And now the big if-else chain
    if (cmd == "world.getBlocks") {
        int x0, y0, z0, x1, y1, z1;
        int ret = sscanf(args.c_str(), "%d,%d,%d,%d,%d,%d", &x0, &y0, &z0, &x1, &y1, &z1);
        if (ret != 6) return fail;

        // Get the blocks
        return getBlocks(commandserver, Vec3{(float) x0, (float) y0, (float) z0}, Vec3{(float) x1, (float) y1, (float) z1});
    } else if (cmd == "world.getPlayerId") {
        for (Player *player : commandserver->minecraft->level->players) {
            if (misc_get_player_username(player) == args)
                return std::to_string(player->id) + "\n";
        }
        WARN("Player [%s] not found.", args.c_str());
        return fail;
    } else if (cmd == "entity.getName") {
        int id;
        int ret = sscanf(args.c_str(), "%d", &id);
        if (ret != 1) return "\n";

        Entity *entity = commandserver->minecraft->level->getEntity(id);
        if (entity == NULL) {
            WARN("Player (or Entity) [%i] not found in entity.getName.", id);
        }
        return misc_get_entity_name(entity) + "\n";
    } else if (cmd == "world.getEntities") {
        int type;
        int ret = sscanf(args.c_str(), "%d", &type);
        if (ret != 1) return "\n";

        std::string result = "";
        for (Entity *entity : commandserver->minecraft->level->entities) {
            if (type == -1 || entity->getEntityTypeId() == type) {
                // Get pos
                // Put it together
                result += getEntityData(commandserver, entity) + "|";
            }
        }
        if (result.size() > 1) result.pop_back();
        return result + "\n";
    } else if (cmd == "world.removeEntity") {
        int id;
        int ret = sscanf(args.c_str(), "%d", &id);
        if (ret != 1) return "0\n";

        Entity *entity = commandserver->minecraft->level->getEntity(id);
        if (entity != NULL && !entity->isPlayer()) {
            entity->remove();
            return "1\n";
        }
        return "0\n";
    } else if (cmd == "world.removeEntities") {
        int type;
        int ret = sscanf(args.c_str(), "%d", &type);
        if (ret != 1) return "0\n";

        int removed = 0;
        for (Entity *entity : commandserver->minecraft->level->entities) {
            if ((type == -1 || entity->getEntityTypeId() == type) && !entity->isPlayer()) {
                entity->remove();
                removed++;
            }
        }
        return std::to_string(removed) + "\n";
    } else if (cmd == "events.chat.posts") {
        std::string ret = "";
        while (chat_events_at != chat_events_start) {
            ChatEvent ce = pop_circular_queue(chatEvents, chat_events_start, chat_events_at);
            std::string message = ce.toString();
            if (compat_mode) std::replace(message.begin(), message.end(), '|', '\\');
            else replace(message, "|", "\\|");
            ret += (compat_mode ? "0," : "") + message + "|";
        }
        if (ret.size() > 1) ret.pop_back();
        return ret + "\n";
    } else if (cmd == "player.events.chat.posts") {
        std::string ret = "";
        while (chat_events_at != chat_events_start) {
            ChatEvent ce = pop_circular_queue(chatEvents, chat_events_start, chat_events_at);
            if (ce.from_player) {
                std::string message = ce.toString();
                if (compat_mode) std::replace(message.begin(), message.end(), '|', '\\');
                else replace(message, "|", "\\|");
                ret += (compat_mode ? "0," : "") + message + "|";
            }
        }
        if (ret.size() > 1) ret.pop_back();
        return ret + "\n";
    } else if (cmd == "events.projectile.hits") {
        std::string result = "";
        while (hit_events_at != hit_events_start) {
            result += pop_circular_queue(hitEvents, hit_events_start, hit_events_at).toString(commandserver) + "|";
        }
        if (result.size() > 1) result.pop_back();
        return result + "\n";
    } else if (cmd == "player.setDirection" && commandserver->minecraft->player) {
        float x, y, z;
        int ret = sscanf(args.c_str(), "%f,%f,%f", &x, &y, &z);
        if (ret != 3) return "";
        setDir((Entity *) commandserver->minecraft->player, x, y, z);
    } else if (cmd == "entity.setDirection") {
        int id;
        float x, y, z;
        int ret = sscanf(args.c_str(), "%d,%f,%f,%f", &id, &x, &y, &z);
        if (ret != 4) return "";
        Entity *entity = commandserver->minecraft->level->getEntity(id);
        if (entity == NULL) {
            WARN("Entity [%i] not found.", id);
        } else {
            setDir(entity, x, y, z);
        }
    } else if (cmd == "player.getDirection" && commandserver->minecraft->player) {
        Vec3 vec = getDir((Entity *) commandserver->minecraft->player);
        return std::to_string(vec.x) + "," + std::to_string(vec.y) + "," + std::to_string(vec.z) + "\n";
    } else if (cmd == "entity.getDirection") {
        int id;
        int ret = sscanf(args.c_str(), "%d", &id);
        if (ret != 1) return "";
        Entity *entity = commandserver->minecraft->level->getEntity(id);
        if (entity == NULL) {
            WARN("Entity [%i] not found.", id);
            return fail;
        } else {
            Vec3 vec = getDir(entity);
            return std::to_string(vec.x) + "," + std::to_string(vec.y) + "," + std::to_string(vec.z) + "\n";
        }
    } else if (cmd == "player.setRotation"  && commandserver->minecraft->player) {
        float yaw;
        int ret = sscanf(args.c_str(), "%f", &yaw);
        if (ret != 1) return "";
        commandserver->minecraft->player->yaw = yaw;
    } else if (cmd == "entity.setRotation") {
        int id;
        float yaw;
        int ret = sscanf(args.c_str(), "%d,%f", &id, &yaw);
        if (ret != 2) return "";
        Entity *entity = commandserver->minecraft->level->getEntity(id);
        if (entity == NULL) {
            WARN("Entity [%i] not found.", id);
        } else {
            entity->yaw = yaw;
        }
    } else if (cmd == "player.setPitch" && commandserver->minecraft->player) {
        float pitch;
        int ret = sscanf(args.c_str(), "%f", &pitch);
        if (ret != 1) return "";
        commandserver->minecraft->player->pitch = pitch;
    } else if (cmd == "entity.setPitch") {
        int id;
        float pitch;
        int ret = sscanf(args.c_str(), "%d,%f", &id, &pitch);
        if (ret != 2) return "";
        Entity *entity = commandserver->minecraft->level->getEntity(id);
        if (entity == NULL) {
            WARN("Entity [%i] not found.", id);
        } else {
            entity->pitch = pitch;
        }
    } else if (cmd == "player.getRotation" && commandserver->minecraft->player) {
        return std::to_string(commandserver->minecraft->player->yaw) + "\n";
    } else if (cmd == "entity.getRotation") {
        int id;
        int ret = sscanf(args.c_str(), "%d", &id);
        if (ret != 1) return "";
        Entity *entity = commandserver->minecraft->level->getEntity(id);
        if (entity == NULL) {
            WARN("Entity [%i] not found.", id);
            return fail;
        } else {
            return std::to_string(entity->yaw) + "\n";
        }
    } else if (cmd == "player.getPitch" && commandserver->minecraft->player) {
        return std::to_string(commandserver->minecraft->player->pitch) + "\n";
    } else if (cmd == "entity.getPitch") {
        int id;
        int ret = sscanf(args.c_str(), "%d", &id);
        if (ret != 1) return "";
        Entity *entity = commandserver->minecraft->level->getEntity(id);
        if (entity == NULL) {
            WARN("Entity [%i] not found.", id);
            return fail;
        } else {
            return std::to_string(entity->pitch) + "\n";
        }
    } else if (cmd == "player.getEntities" || cmd == "entity.getEntities") {
        // Parse
        int dist, type;
        Entity *src = NULL;
        if (cmd == "player.getEntities") {
            int ret = sscanf(args.c_str(), "%d,%d", &dist, &type);
            if (ret != 2 || commandserver->minecraft->player == NULL) return "";
            src = (Entity *) commandserver->minecraft->player;
        } else {
            int id = 0;
            int ret = sscanf(args.c_str(), "%d,%d,%d", &id, &dist, &type);
            if (ret != 3) return "";
            src = commandserver->minecraft->level->getEntity(id);
            if (src == NULL) {
                WARN("Entity [%i] not found.", id);
                return fail;
            }
        }
        // Run
        std::string result = "";
        for (Entity *entity : commandserver->minecraft->level->entities) {
            if ((type == -1 || entity->getEntityTypeId() == type) && distance_between(src, entity) < dist) {
                result += getEntityData(commandserver, entity) + "|";
            }
        }
        if (result.size() > 1) result.pop_back();
        return result + "\n";
    } else if (cmd == "player.removeEntities" || cmd == "entity.removeEntities") {
        // Parse
        int dist, type;
        Entity *src = NULL;
        if (cmd == "player.removeEntities") {
            int ret = sscanf(args.c_str(), "%d,%d", &dist, &type);
            if (ret != 2 || commandserver->minecraft->player == NULL) return "";
            src = (Entity *) commandserver->minecraft->player;
        } else {
            int id = 0;
            int ret = sscanf(args.c_str(), "%d,%d,%d", &id, &dist, &type);
            if (ret != 3) return "";
            src = commandserver->minecraft->level->getEntity(id);
            if (src == NULL) {
                WARN("Entity [%i] not found.", id);
                return fail;
            }
        }
        // Run
        int removed = 0;
        for (Entity *entity : commandserver->minecraft->level->entities) {
            if ((type == -1 || entity->getEntityTypeId() == type) && distance_between(src, entity) < dist && !entity->isPlayer()) {
                entity->remove();
                removed++;
            }
        }
        return std::to_string(removed) + "\n";
    } else if (cmd == "world.setSign") {
        // Parse
        int x, y, z, id, data;
        char l1[100], l2[100], l3[100], l4[100];
        int ret = sscanf(args.c_str(), "%d,%d,%d,%d,%d,%99[^,],%99[^,],%99[^,],%99s", &x, &y, &z, &id, &data, l1, l2, l3, l4);
        if (ret < 5) return fail;
        // Translate
        commandserver->pos_translator.from(x, y, z);
        // Set block
        commandserver->minecraft->level->setTileAndData(x, y, z, id, data);
        // Set sign data
        if (ret == 3) return "";
        SignTileEntity *sign = (SignTileEntity *) commandserver->minecraft->level->getTileEntity(x, y, z);
        if (sign == NULL || sign->type != 4) return "";
        if (ret > 5) sign->lines[0] = l1;
        if (ret > 6) sign->lines[1] = l2;
        if (ret > 7) sign->lines[2] = l3;
        if (ret > 8) sign->lines[3] = l4;
    } else if (cmd == "world.spawnEntity") {
        // Parse
        float x, y, z;
        int id;
        int ret = sscanf(args.c_str(), "%f,%f,%f,%d", &x, &y, &z, &id);
        if (ret != 4) return fail;
        // Translate
        int ix = x, iy = y, iz = z;
        int ix1 = x, iy1 = y, iz1 = z;
        commandserver->pos_translator.from(ix, iy, iz);
        x += ix - ix1;
        y += iy - iy1;
        z += iz - iz1;
        if (compat_mode && modern_entity_id_mapping.find(id) != modern_entity_id_mapping.end()) id = modern_entity_id_mapping[id];
        // Spawn
        Entity *entity = NULL;
        if (id < 0x40) {
            entity = (Entity *) MobFactory::CreateMob(id, commandserver->minecraft->level);
        } else {
            entity = EntityFactory::CreateEntity(id, commandserver->minecraft->level);
        }
        if (entity == NULL) return fail;
        entity->moveTo(x, y, z, 0, 0);
        commandserver->minecraft->level->addEntity(entity);
        return std::to_string(entity->id) + "\n";
    } else if (cmd == "world.getEntityTypes") {
        std::string result = "";
        for (auto &i : misc_get_entity_names()) {
            if (compat_mode && modern_entity_id_mapping.find(i.first) != modern_entity_id_mapping.end()) result += std::to_string(modern_entity_id_mapping[i.first]);
            else result += std::to_string(i.first);
            result += "," + i.second + "|";
        }
        if (result.size() > 1) result.pop_back();
        return result + "\n";
    } else if (cmd == "player.setAbsPos" && commandserver->minecraft->player) {
        float x, y, z;
        int ret = sscanf(args.c_str(), "%f,%f,%f", &x, &y, &z);
        if (ret != 3) return fail;
        commandserver->minecraft->player->moveTo(x, y, z, commandserver->minecraft->player->yaw, commandserver->minecraft->player->pitch);
    } else if (cmd == "player.getAbsPos" && commandserver->minecraft->player) {
        return std::to_string(commandserver->minecraft->player->x) + ","
             + std::to_string(commandserver->minecraft->player->y) + ","
             + std::to_string(commandserver->minecraft->player->z) + "\n";
    } else if (cmd == "reborn.disableCompatMode") {
        compat_mode = false;
    } else if (cmd == "reborn.enableCompatMode") {
        compat_mode = true;
    } else {
        // Either invalid or a vanilla command, either way hand it off to the orignal handler
        return old(commandserver, client, command);
    }
    // No return, it's different from empty! Empty is "\n". "" will result in nothing being sent back while "\n" will result in an empty string being sent back
    return "";
}

// Arrow entity hit
static Entity *shooter = NULL;
static HitResult *Arrow_tick_HitResult_constructor_injection(HitResult *self, Entity *target) {
    // Orignal
    self->type = 1;
    self->entity = target;
    self->unknown = false;
    self->exact.x = target->x;
    self->exact.y = target->y;
    self->exact.z = target->z;
    // Add event
    if (shooter && shooter->isPlayer()) {
        push_circular_queue(ProjectileHitEvent{
            .x = (int) target->x,
            .y = (int) target->y,
            .z = (int) target->z,
            .owner = misc_get_player_username((Player *) shooter),
            .targetId = target->id
        }, hitEvents, hit_events_start, hit_events_at);
    }
    return self;
}

// Arrow block hit
static void Arrow_tick_injection(Arrow_tick_t old, Arrow *self) {
    int oldFlightTime = self->flight_time;
    shooter = self->level->getEntity(self->getAuxData());
    old(self);
    if (self && !self->pending_removal && self->grounded && oldFlightTime != self->flight_time) {
        if (shooter && shooter->isPlayer()) {
            // Hit! Get the data
            push_circular_queue(ProjectileHitEvent{
                .x = self->hit_x,
                .y = self->hit_y,
                .z = self->hit_z,
                .owner = misc_get_player_username((Player *) shooter),
                .targetId = 0
            }, hitEvents, hit_events_start, hit_events_at);
        }
    }
}

// Throwable hits
static void Throwable_tick_Throwable_onHit_injection(Throwable *self, HitResult *res) {
    if (res == NULL || res->type == 2) return self->onHit(res);
    Entity *thrower = self->level->getEntity(self->getAuxData());
    if (thrower == NULL || !thrower->isPlayer()) return self->onHit(res);
    ProjectileHitEvent event;
    if (res->type == 1 && res->entity) {
        // Entity
        event = ProjectileHitEvent {
            .x = (int) res->exact.x,
            .y = (int) res->exact.y,
            .z = (int) res->exact.z,
            .owner = misc_get_player_username((Player *) thrower),
            .targetId = res->entity->id
        };
    } else {
        // Tile
        event = ProjectileHitEvent {
            .x = res->x,
            .y = res->y,
            .z = res->z,
            .owner = misc_get_player_username((Player *) thrower),
            .targetId = 0
        };
    }
    push_circular_queue(event, hitEvents, hit_events_start, hit_events_at);
    self->onHit(res);
}

static void Gui_addMessage_injection(Gui_addMessage_t original, Gui *gui, const std::string &text) {
    push_circular_queue(ChatEvent{
        text,
        is_sending_to_chat()
    }, chatEvents, chat_events_start, chat_events_at);
    original(gui, text);
}

void init_api() {
    if (feature_has("Implement RaspberryJuice API", server_enabled)) {
        overwrite_calls(CommandServer_parse, CommandServer_parse_injection);
        overwrite_calls(Arrow_tick, Arrow_tick_injection);
        overwrite_call((void *) 0x8b1e8, (void *) Arrow_tick_HitResult_constructor_injection);
        overwrite_call((void *) 0x8c5a4, (void *) Throwable_tick_Throwable_onHit_injection);
        overwrite_calls(Gui_addMessage, Gui_addMessage_injection);
    }
}
